
import sys as _sys
import os as _os
import string as _string

class BlockDev:
    def get_major(self):
        return self._BlockDev__device.major
    def set_major(self, major):
        self._BlockDev__device.major = major
    major = property(get_major, set_major, None, "major device number")

    def get_minor(self):
        return self.__device.minor
    def set_minor(self, minor):
        self._BlockDev__device.minor = minor
    minor = property(get_minor, set_minor, None, "minor device number")

    def get_device(self):
        return self._BlockDev__device
    device = property(get_device, None, None, "device number")

    def get_dmdev(self):
        return "%s:%s" % (self._BlockDev__device.major, self._BlockDev__device.minor)
    dmdev = property(get_dmdev, None, None, "device number formatted for dm")

    def get_path(self):
        return self._BlockDev__path
    def set_path(self, path):
        self._BlockDev__path = str(path)
    path = property(get_path, set_path, None, "path to device node")

    def get_mode(self):
        return self._BlockDev__device.mode
    def set_mode(self, mode):
        self._BlockDev__device.mode = int(mode)
    mode = property(get_mode, set_mode, None, "mode for device node")

    def get_scsi_id(self):
        if self._BlockDev__scsi_id:
            return self._BlockDev__scsi_id

        if not self.path:
            raise RuntimeError, "path not set"
        
        try:
            self._BlockDev__scsi_id = bdevid.scsi_id(self.path)
        except ValueError:
            self._BlockDev__scsi_id = None
        return self._BlockDev__scsi_id
    id = property(get_scsi_id, None, None, "unique id for device")

    def __init__(self, path=None, major=None, minor=None, dev=None):
        self._BlockDev__context = None
        self._BlockDev__scsi_id = None

        self._BlockDev__path = None
        if not path is None:
            self.FromFile(path)
        elif not (major is None and minor is None):
            self.FromMajorMinor(major, minor)
        elif not dev is None:
            self.FromDev(dev)
        else:
            import dm as _dm
            self._BlockDev__device = _dm.device(0,0)
            del _dm

        # self.sysfs = None # XXX not handling yet
        # XXX should probe parents/children/etc
        self.parents = []
        self.children = []

        self.group = None
        self.group_position = None
        self.group_siblings = []

    def FromFile(self, path):
        import dm as _dm

        path = str(path)
        self._BlockDev__path = path
        self._BlockDev__device = _dm.device(path=path)
        del _dm
        return self;

    def FromMajorMinor(self, major, minor):
        import dm as _dm

        self._BlockDev__device = _dm.device(major, minor)
        del _dm
        return self

    def FromDev(self, dev):
        import dm as _dm

        self._BlockDev__device == _dm.device(0,0)
        self._BlockDev__device.dev = dev
        del _dm
        return self

    def create(self, path=None, mode=None, context=None):
        args = {}
        if path is None:
            path = self.path
        if path is None:
            raise ValueError, "no path set for %s" % (self,)
        args['path'] = path
        args['mode'] = self.mode
        if not mode is None:
            args['mode'] = mode
        if context is None:
            context = self._BlockDev__context
        if not context is None:
            args['context'] = context

        return apply(self.device.mknod, (), args)

    def output(self, *args):
        for x in args:
            print x

    def remove(self):
        if not self.path:
            raise RuntimeError, "no path set for %s" % (self,)
        if self.path.split('/')[-1].startswith('VolGroup'):
            raise RuntimeError, "trying to unlink %s" % (self.path,)
        _os.unlink(self.path)

class Device:
    def get_major(self):
        return self.__bdev.major
    def set_major(self, major):
        self.__bdev.major = major
    major = property(get_major, set_major, None, "major device number")

    def get_minor(self):
        return self.__bdev.minor
    def set_minor(self, major):
        self.__bdev.minor = major
    minor = property(get_major, set_major, None, "major device number")

    def get_dmdev(self):
        return self.__bdev.dmdev
    dmdev = property(get_dmdev, None, None, "device number formatted for dm")

    def get_map(self):
        return self.__map
    def set_map(self, map):
        # XXX type chek this
        self.__map = map
    map = property(get_map, set_map, None, "device map for this device")

    def __init__(self, bdev=BlockDev(), map=None):
        if not isinstance(bdev, BlockDev):
            raise TypeError, "bdev must be an instance of block.BlockDev"
        self.__bdev = bdev
        self.__map = map

    def FromFile(self, path):
        self.__bdev = BlockDev().FromFile(path)
        self.__map = None

class RaidSet:
    def __init__(self, rs, prefix="/dev/mapper/"):
        self.rs = rs
        self._RaidSet__prefix = prefix
        self._RaidSet__bdev = None
        self._RaidSet__map = None
        self._RaidSet__parts = []
        self._RaidSet__name = rs.name
        self.active = False
        self.mknod = False

    def __cmp__(self, other):
        return cmp(self.map, other.map)

    def get_prefix(self):
        return self._RaidSet__prefix
    def set_prefix(self, value):
        self._RaidSet__prefix = value
    prefix = property(get_prefix, set_prefix, None, \
        "prefix for path to device nodes")
        
    def get_PedDevice(self):
        import parted as _parted
        ret = _parted.PedDevice.get(self.bdev.path)
        del _parted
        return ret
    PedDevice = property(get_PedDevice, None, None, "parted.PedDevice")

    def get_level(self):
        if self.rs.dmtype in ("stripe", "striped"):
            return 0
        elif self.rs.dmtype == "mirror":
            return 1
        raise NotImplementedError, "unknown dmtype %s" % (rs.dmtype,)
    level = property(get_level, None, None, "raid level")

    def get_members(self, descend=True):
        import dmraid as _dmraid

        for c in self.rs.children:
            if isinstance(c, _dmraid.raidset):
                r = RaidSet(c, prefix=self.prefix)
                yield r
                if descend:
                    for m in r.members:
                        yield m
            elif isinstance(c, _dmraid.raiddev):
                yield RaidDev(c)

        del _dmraid

    members = property(get_members, None, None, "members")

    def get_member_devpaths(self, descend=True):
        for m in self.get_members(descend):
            if isinstance(m, RaidDev):
                yield m.devpath
    member_devpaths = property(get_member_devpaths, None, None, "member devpaths")

    def get_spares(self, descend=True):
        import dmraid as _dmraid

        for s in self.rs.spares:
            if isinstance(c, _dmraid.raidset):
                r = RaidSet(c)
                yield r
                if descend:
                    for m in r.members:
                        yield m
                    for m in r.spares:
                        yield m
            elif isinstance(c, _dmraid.raiddev):
                yield RaidDev(c)
        # members might have spares as well
        for c in self.rs.children:
            if isinstance(c, _dmraid.raidset) and descend:
                for spare in RaidSet(c).spares:
                    yield spare

        del _dmraid
    spares = property(get_spares, None, None, "spares")

    def get_bdev(self):
        if not self._RaidSet__bdev is None:
            return self._RaidSet__bdev
        try:
            self._RaidSet__bdev = BlockDev(self.prefix + self.name)
            return self._RaidSet__bdev
        except:
            pass
    bdev = property(get_bdev, None, None, "block.BlockDev")

    def get_map(self):
        if not self._RaidSet__map is None:
            return self._RaidSet__map

        # we get "/dev/hda" from one and "3:0" from the other, so we have to
        # fix up the device name
        def map_dev(path):
            if path[0] != '/':
                return path.strip()

            try:
                newpath=path.split('/')[-1]
                pos = 0
                dev = None 
                num = None
                while pos < len(newpath):
                    if newpath[pos].isdigit():
                        dev = newpath[:pos]
                        num = newpath[pos:]
                        break
                    pos += 1
                if dev is None:
                    dev = newpath
                sysnewpath = None
                if num is None:
                    sysnewpath = '/sys/block/%s/dev' % (dev,)
                else:
                    sysnewpath = '/sys/block/%s/%s%s/dev' % (dev, dev, num)
                f = open(sysnewpath, 'r')
                l = f.readline()
                return l.strip()
            except:
                return path.strip()

        parts = str(self.rs.dmTable).split(' ')

        table = []
        for part in parts:
            table += [map_dev(part),]
        table = _string.join(table, ' ')

        import dm as _dm

        for map in _dm.maps():
            # XXX wtf?  why's it have a space at the end sometimes?
            if str(map.table).strip() == table:
                self._RaidSet__map = map
                self.active = True
                del _dm
                return self._RaidSet__map

        # all else has failed, make an ew map...
        self._RaidSet__map = _dm.map(name=self.name, table=self.rs.dmTable)
        self.active = True
        del _dm
        return self._RaidSet__map
    map = property(get_map, None, None, "block.dm.map")

    def get_name(self):
        return self._RaidSet__name
    def set_name(self, name):
        bdev = self._RaidSet__bdev
        prefix = self.prefix

        self.deactivate()
        self.rs.name = name
        self._RaidSet__name = name
        self._RaidSet__bdev = BlockDev().FromMajorMinor(bdev.major, bdev.minor)
        self.bdev.mode = 0600
        self.bdev.path = prefix + name
        self.activate()

    name = property(get_name, set_name, None, "this name of this RaidSet")

    def get_parts(self):
        for x in self._RaidSet__parts:
            yield x
    partitions = property(get_parts, None, None, "this device's partitions")

    def get_valid(self):
        if self.rs.degraded:
            return False
        for x in self.members:
            if isinstance(x, RaidDev) or isinstance(x, RaidSet):
                if not x.valid:
                    return False
        return True
    valid = property(get_valid, None, None, "test a raidset for validity")

    def output(self, *args):
        for x in args:
            print x

    def buildParts(self):
        from maps import PartitionDeviceMap as _PartitionDeviceMap
        import parted as _parted

        dev = self.PedDevice
        dev.open()
        try:
            disk = _parted.PedDisk.new(dev)
        except _parted.error, msg:
            return
        part = disk.next_partition()
        while part:
            if part.is_active():
                name = "%sp%s" % (self.name, part.num)
                bdev = BlockDev(self.prefix + self.name)
                bdev.mode = 0600
                map = _PartitionDeviceMap(0, part.geom.length, bdev,
                    part.geom.start, part.num)
                map.name = name
                map.create()
                self._RaidSet__parts.append(map)
            part = disk.next_partition(part)
        del disk
        dev.close()
        del _PartitionDeviceMap
        del _parted

    def removeMemberParts(self):
        for m in self.members:
            if isinstance(m, RaidDev):
                m.removeParts()
            elif isinstance(m, RaidSet):
                m.removeMemberParts()

    def scanMemberParts(self):
        for m in self.members:
            if isinstance(m, RaidDev):
                m.scanParts()
            elif isinstance(m, RaidSet):
                m.scanMemberParts()

    def activate(self, degradedOk=False, mknod=False):
        if self.active:
            return
        self.removeMemberParts()
        if mknod:
            try:
                _os.unlink(self.prefix+self.name)
            except:
                pass
            self.map.dev.mknod(self.prefix+self.name)
            self.mknod = True

        self.buildParts()
        ret = self.partitions
        return ret

    def deactivate(self):
        if not self.active:
            return
        parts = self._RaidSet__parts
        self._RaidSet__parts = []
        for part in parts:
            part.map.remove()
        parts = list(self.partitions)
        if len(parts) == 0 and not self._RaidSet__map is None:
            self.map.remove()
            self.active = False
            self._RaidSet__map = None
        elif len(parts) > 0:
            raise RuntimeError, "raidset has active partitions"
        self.scanMemberParts()

    def display(self, space=0, printer=lambda x,y: _sys.stdout.write("%s%s\n" % (x*' ',y))):
        printer(space, self)
        for m in self.members:
            m.display(space+1, printer=printer)

class RaidDev:
    def __init__(self, rd, prefix=None):
        self.rd = rd
        self._RaidDev__bdev = None
        self._RaidDev__prefix = None

    def get_prefix(self):
        if self._RaidDev__prefix is None:
            self._RaidDev__prefix = \
                _string.join(self.rd.device.path.split('/')[:-1],'/') + '/'
        return self._RaidDev__prefix
    prefix = property(get_prefix, None, None, "prefix for device node path")

    def get_bdev(self):
        if self._RaidDev__bdev is None:
            bdev = BlockDev().FromFile(self.rd.device.path)
            self._RaidDev__bdev = bdev
        return self._RaidDev__bdev
    bdev = property(get_bdev, None, None, "block.BlockDev")

    def get_PedDevice(self):
        import parted as _parted
        ret = _parted.PedDevice.get(self.bdev.path)
        del _parted
        return ret
    PedDevice = property(get_PedDevice, None, None, "parted.PedDevice")

    def get_devpath(self):
        return self.rd.device.path.split('/')[-1]
    devpath = property(get_devpath, None, None, "path relative to /dev")

    def get_valid(self):
        if self.bdev is not None and \
                hasattr(self.bdev, 'path') and self.bdev.path is not None:
            return True
        return False
    valid = property(get_valid, None, None, "test a raiddev for validity")

    def __cmp__(self, other):
        # Dear Python, you are retarded.
        def strcmp(a,b):
            ab = zip(reduce(lambda x,y: x + [ord(y)], list(a), []),
                reduce(lambda x,y: x + [ord(y)], list(b), []))
            cmps = filter(None, map(lambda (x,y): cmp(x,y), ab)) or \
                [len(a) - len(b),]
            return cmps[0]

        obdev = other
        if isinstance(other, "".__class__):
            filename = '/dev/' + other
            if other[0].isdigit():
                maj,min = (0,0)
                if ':' in other:
                    maj,min = other.split(':')[0:2]
                    obdev = BlockDev().FromMajorMinor(maj,min)
                else:
                    device = str(self.bdev.device)
                    return strcmp(device, other)
            elif other.startswith("/"):
                try:
                    obdev = BlockDev().FromFile(other)
                except:
                    filename = self.devpath
            elif other.startswith('/dev/mapper/'):
                filename = '/dev/mapper/' + self.devpath
            elif other.startswith('/dev/'):
                filename = '/dev/' + self.devpath
            elif other.startswith('/tmp/'):
                filename = '/tmp/' + self.devpath

            try:
                obdev = BlockDev().FromFile(filename)
            except:
                return strcmp(filename, other)

        if isinstance(obdev, BlockDev):
            try:
                return cmp(self.bdev.device.dev, obdev.device.dev)
            except:
                pass

        elif isinstance(other, RaidDev):
            if self.rd.device.serial is not None and \
                    other.rd.device.serial is not None:
                return strcmp(self.rd.device.serial, other.rd.device.serial)

            elif self.bdev is not None and other.bdev is not None:
                return strcmp(self.bdev.device.dev, other.bdev.device.dev)

            else:
                return strcmp(self.devpath, other.devpath)
        return -1 

    def removeParts(self):
        for x in range(1,257):
            self.rd.device.rmpart(x)

    def scanParts(self):
        self.rd.device.scanparts()

    def display(self, space=0, printer=lambda x,y: _sys.stdout.write("%s%s\n" % (x*' ',y))):
        printer(space, self)
